/*
 * Copyright (c) 2001, Michael Alyn Miller <malyn@strangeGizmo.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Michael Alyn Miller nor the names of the
 *    contributors to this software may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package com.strangegizmo.filter;

/* Java imports. */
import java.io.*;
import java.util.*;
import java.util.zip.*;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * The GzipFilter processes requests with an Accept-Encoding of `gzip'
 * and compresses the data.
 *
 * @author		Michael Alyn Miller <malyn@strangeGizmo.com>
 * @version		1.0.1
 */
public class GzipFilter implements Filter {
	/** The attribute that is set on the ServletRequest object to
	 * indicate that the GzipFilter is in use for this request. */
	public static final String inUseAttributeName_
		= "com.strangegizmo.filter.GzipFilter.inUse";


	/**
	 * Initializes the filter.
	 */
	public void init(FilterConfig filterConfig) {}

	/**
	 * Destroys the filter.
	 */
	public void destroy() {}

	/**
	 * Process a new Request/Response pair, wrapping the Response in a
	 * GzipFilterResponse if the response data can be compresses.
	 */
	public void doFilter(
		ServletRequest request, ServletResponse response, FilterChain chain)
		throws IOException, ServletException
	{
		/* Only compress HTTP requests. */
		if (!(request instanceof HttpServletRequest)
			|| !(response instanceof HttpServletResponse))
		{
			chain.doFilter(request, response);
			return;
		}

		/* Verify that we have not already seen this request. */
		if (request.getAttribute(inUseAttributeName_) != null) {
			chain.doFilter(request, response);
			return;
		}


		/* Retrieve the Accept-Encoding header and look for the gzip
		 * value. */
		Enumeration e
			= ((HttpServletRequest)request).getHeaders("Accept-Encoding");
		boolean gzipOk = false;
		while (e.hasMoreElements()) {
			/* Look for `gzip' in this value, exiting the loop if we
			 * find it. */
			String value = (String)e.nextElement();
			if (value.indexOf("gzip") != -1) {
				gzipOk = true;
				break;
			}
		}

		/* Only consider requests where Accept-Encoding includes gzip. */
		if (!gzipOk) {
			chain.doFilter(request, response);
			return;
		}


		/* Add our inUse attribute to the request. */
		request.setAttribute(inUseAttributeName_, "true");

		/* Wrap the response object in a GzipFilterResponse and continue
		 * the filter chain. */
		GzipFilterResponseWrapper responseWrapper
			= new GzipFilterResponseWrapper((HttpServletResponse)response);
		chain.doFilter(request, responseWrapper);

		/* Finish the output stream. */
		responseWrapper.finish();
	}
}


/**
 * The GzipFilterResponseWrapper extends HttpServletResponseWrapper to
 * provide Gzip-wrapped OutputStreams and Writers.
 *
 * @author		Michael Alyn Miller <malyn@strangeGizmo.com>
 * @version		1.0.1
 */
class GzipFilterResponseWrapper extends HttpServletResponseWrapper {
	/** The gzip-wrapped ServletOutputStream. */
	protected GzipFilterOutputStream out_ = null;

	/** The gzip-wrapped PrintWriter. */
	protected PrintWriter writer_ = null;


	/**
	 * Creates a new response wrapper.
	 *
	 * @param response The wrapped response.
	 */
	GzipFilterResponseWrapper(HttpServletResponse response) {
		super(response);
	}


	/**
	 * Flushes the buffer to the client.
	 */
	public void flushBuffer() throws IOException {
		if (writer_ != null)
			writer_.flush();
		if (out_ != null)
			out_.flush();
	}

	/**
	 * Finishes the gzip-wrapped response by completely flushing the
	 * output stream and finishing the gzip encoding.
	 */
	public void finish() throws IOException {
		if (out_ != null)
			out_.finish();
	}

	/**
	 * Returns a (gzip-wrapped) output stream for writing to the client.
	 * Either this method or getWriter() may be called to write to the
	 * client, but not both.
	 *
	 * @return A (gzip-wrapped) output stream for writing to the client.
	 */
	public ServletOutputStream getOutputStream() throws IOException {
		/* Throw an exception if a PrintWriter has been created. */
		if (writer_ != null)
			throw new IllegalStateException(
				"Cannot call getOutputStream() after getWriter().");

		/* Create our gzip-wrapped ServletOutputStream if we do not have
		 * one already. */
		if (out_ == null)
			out_ = new GzipFilterOutputStream(
				(HttpServletResponse)getResponse());

		/* Return the gzip-wrapped ServletOutputStream. */
		return out_;
	}

	/**
	 * Returns a (gzip-wrapped) PrintWriter for writing to the client.
	 * Either this method or getOutputStream() may be called to write to
	 * the client, but not both.
	 *
	 * @return A (gzip-wrapped) PrintWriter for writing to the client.
	 */
	public PrintWriter getWriter() throws IOException {
		/* Throw an exception if a ServletOutputStream has been created. */
		if (out_ != null)
			throw new IllegalStateException(
				"Cannot call getWriter() after getOutputStream().");

		/* Create our gzip-wrapped PrintWriter if we do not have one
		 * already. */
		if (writer_ == null)
			writer_ = new PrintWriter(getOutputStream());

		/* Return the gzip-wrapped PrintWriter. */
		return writer_;
	}
}


/**
 * The GzipFilterOutputStream extends ServletOutputStream to output
 * Gzip-compressed data.
 *
 * @author		Michael Alyn Miller <malyn@strangeGizmo.com>
 * @version		1.0.1
 */
class GzipFilterOutputStream extends ServletOutputStream {
	/** The GZIPOutputStream to pass data through. */
	protected GZIPOutputStream out_ = null;


	/**
	 * Creates a new GzipFilterOutputStream.
	 */
	GzipFilterOutputStream(HttpServletResponse response) throws IOException {
		/* Initialize the superclass. */
		super();

		/* Create the GZIPOutputStream. */
		out_ = new GZIPOutputStream(response.getOutputStream());

		/* Add the Content-Encoding header to the response. */
		response.addHeader("Content-Encoding", "gzip");
	}


	/**
	 * Closes the stream.
	 */
	public void close() throws IOException {
		/* Flush the stream. */
		flush();

		/* Close the output stream. */
		out_.close();
	}

	/**
	 * Flushes the stream.
	 */
	public void flush() throws IOException {
		/* Flush the output stream. */
		out_.flush();
	}

	/**
	 * Finishes the stream by completely flushing and closing the gzip
	 * filter.  Unlike flush(), this will cause all further writes to
	 * the stream to fail.
	 */
	public void finish() throws IOException {
		/* Flush first. */
		flush();

		/* Finish the GZIPOutputStream, otherwise it will not write out
		 * all of the compressed data. */
		out_.finish();
	}


	/**
	 * Writes a byte of data to the stream.
	 *
	 * @param b The byte.
	 */
	public void write(int b) throws IOException {
		out_.write(b);
	}

	/**
	 * Writes an array of bytes to the stream.
	 *
	 * @param b The array of bytes.
	 */
	public void write(byte[] b) throws IOException {
		out_.write(b);
	}

	/**
	 * Writes an array of bytes to the stream, given an offset into the
	 * array at which to start writing and a count of the number of
	 * bytes to write.
	 *
	 * @param b The array of bytes.
	 * @param off The offset from which to read.
	 * @param len The count of bytes to read at <code>off</code>.
	 */
	public void write(byte[] b, int off, int len) throws IOException {
		out_.write(b, off, len);
	}
}
